#include "../common/shared.h"
#include "../common/serverlib.h"
#include "ipc_shared.h"

#include <signal.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <setjmp.h>

struct timeRequest our_timeRequest; /*holds our time request*/
char * reply_string = NULL; /*Refers to the reply shared memory segment*/
struct timeRequest* server_ptimeRequest = NULL; /*refers to the servers` request shared memory segment*/
int serverSemaphoreID_int = -1; /*server's semaphore ID*/
sigjmp_buf interrogation_sigjmpBuf; /*used for sigsetjmp and siglongjmp*/
volatile boolean canjump_boolean = FALSE;
volatile int numInterrogations_int = 0;
const int MAXINTERROGATIONS = 20;

/*
function setUpCommunication
task: connect to server's IPC and initialize client's
input: none
output: -1 on error, 0 otherwise
*/
int setUpCommunication(){
  int serverSharedMemSegmentID_int;
  /*==========  Connect to IPC structures of server =========*/
  serverSemaphoreID_int = semget(SERVER_SEMAPHORES_KEY,2,IPC_CONNECT_FLAG);
  if (serverSemaphoreID_int == -1){
    WriteError("Cannot get semaphore ID of server\n");
    return -1;
  }

  serverSharedMemSegmentID_int = shmget(SERVER_SHARED_MEMORY_KEY,REQUESTSIZE,IPC_CONNECT_FLAG);
  if(serverSemaphoreID_int == -1){
    WriteError("Cannot get server shared memory\n");
    return -1;
  }

  server_ptimeRequest = shmat(serverSharedMemSegmentID_int, NULL, 0);

  if (server_ptimeRequest == (struct timeRequest *) -1) {
    perror("shmat");
    WriteError("Cannot attach server's shared memory\n");
    return -1;
  }



  /*========== Create IPC structures for client*/
  our_timeRequest.replyMemorySegmentID_int = shmget(IPC_PRIVATE, MAXREPLYSIZE, IPC_CREATION_FLAG);
  if (our_timeRequest.replyMemorySegmentID_int  == -1) {
    WriteError("Cannot initialize client-side shared memory\n");
    return -1;
  }

  reply_string = shmat(our_timeRequest.replyMemorySegmentID_int, NULL, 0);
  if (reply_string == (char *) -1){
    WriteError("Cannot attach shared memory of client\n");
    return -1;
  }

  our_timeRequest.replyNotificationSemaphore_int =semget(IPC_PRIVATE, 1, IPC_CREATION_FLAG);
  if (our_timeRequest.replyNotificationSemaphore_int == -1) {
    WriteError("Cannot get semaphore for client\n");
    server_ptimeRequest = NULL;
    return -1;
  }



  return 0;
}

/*
function tearDownCommunication
task: destroy client-specific IPC structures
input: none
output: none
*/
void tearDownCommunication(){
  /*Release link to server's shared memory */
  if (server_ptimeRequest != NULL) shmdt(server_ptimeRequest);
/*Clean semaphore*/
  if(our_timeRequest.replyNotificationSemaphore_int  != -1) {
    semctl(our_timeRequest.replyNotificationSemaphore_int , 0, IPC_RMID, 0); /*Delete semaphore set*/
    our_timeRequest.replyNotificationSemaphore_int  = -1;
  }

  /*Clean shared memory segment*/
  if(our_timeRequest.replyMemorySegmentID_int != -1) {
   /*Detach memory segment*/
    shmdt(reply_string);
    reply_string  = NULL;

    /*remove shared memory segment from memory*/
    shmctl(our_timeRequest.replyMemorySegmentID_int, IPC_RMID, NULL);
    our_timeRequest.replyMemorySegmentID_int = -1;
  }

}
/*
function SignalHandler
task: catch SIGALRM and terminating signal
        For terminating signals, clean IPC and exit
	For alarms, jump to related segment for processing
*/
static void SignalHandler(int sig){
  if (sig != SIGALRM){
    WriteError("time client: terminating on signal\n");
    tearDownCommunication();
    exit(1);
  }
  else{
	if (canjump_boolean){
	  if (numInterrogations_int < MAXINTERROGATIONS)  siglongjmp(interrogation_sigjmpBuf, 1);
          else {
	      tearDownCommunication();
              exit(0);
	  }
	}
  }
}

/*
function InitSignalHandler
task: initialize the signal handler for terminating signals and alarms to SignalHandler
input: none
output: -1 on error, 0 otherwise
*/
int InitSignalHandler(){
  struct sigaction settings_sigaction;

	settings_sigaction.sa_flags = SA_NOCLDSTOP; /*Notify that we don't care for stopped child*/
	settings_sigaction.sa_handler = SignalHandler;
	/*Will ignore all signals when handler is on*/
	if (sigfillset(&settings_sigaction.sa_mask) == -1) return -1;


	if(sigaction(SIGALRM,&settings_sigaction, NULL ) == -1) return -1;
        if (setSignalHandlerForTerminatingSignals(settings_sigaction) == -1) return -1;

	return 0;

}

/*
function forwardRequest
task: obtain the SERVER_READY lock and put the local time request to the shared memory segment
input: none
output: -1 on error, 0 otherwise
*/
int forwardRequest(){
  struct sembuf lock_sembuf;

  if (server_ptimeRequest == NULL) return -1;

  /*get lock*/
  lock_sembuf.sem_num =SERVER_READY;
  lock_sembuf.sem_op = -1;
  lock_sembuf.sem_flg = SEM_UNDO;
  if (semop(serverSemaphoreID_int, &lock_sembuf, 1) == -1) return -1;

    /*send data*/
  memcpy(server_ptimeRequest,&our_timeRequest,sizeof(our_timeRequest));

  /*Notify*/
  lock_sembuf.sem_num =REQUEST_FORWARDED_NOTIFICATION;
  lock_sembuf.sem_op = +1;
/*  lock_sembuf.sem_flg = 0;*/
lock_sembuf.sem_flg = SEM_UNDO;
  if (semop(serverSemaphoreID_int, &lock_sembuf, 1) == -1) return -1;

  return 0;
}

/*
function serverReply
task: return the server's reply to the request
input: none
output: pointer to a zero-terminated string holding the reply
*/
char * serverReply(){
 struct sembuf lock_sembuf;
  /*get lock*/
  lock_sembuf.sem_num =0;
  lock_sembuf.sem_op = -1;
  /*lock_sembuf.sem_flg = 0;*/
  lock_sembuf.sem_flg = SEM_UNDO;
  if (semop(our_timeRequest.replyNotificationSemaphore_int, &lock_sembuf, 1) == -1) return NULL;

  /*return shared memory*/
  return reply_string;
}

/*
Main method
*/
int main(int argc, char**argv){
  const int ALARM_DEFAULT = 60;
  char * reply_sz;
  char outputMessage_sz[1024];
  pid_t client_pidT = getpid();
  boolean sendRequest_boolean = FALSE;
  sigset_t criticalSectionSigMask_sigsetT;
  sigset_t normalSigMask_sigsetT;
  char * elapsedTimeInfo_sz;
  boolean firstRun_boolean = TRUE;

  /*===================================Initialization==================*/
  void initializeClock();

  if (setUpCommunication() == -1){
    tearDownCommunication();
    WriteError("time client: Error in setting IPC\n");
    exit(-1);
  }

  if (InitSignalHandler() == -1){
    tearDownCommunication();
    WriteError("time client: Error in setting Signal handling routine\n");
    exit(-1);
  }

  if (ensureStandardOutputAtomicity() == -1){
    tearDownCommunication();
    WriteError("Unable to ensure output atomicity\n");
  }
  /*Initialize request*/
  our_timeRequest.reqType = TIME;
  our_timeRequest.clientID = getpid();

WriteError("Handlers and IPC Set\n");

  /*=== Prepare alarms*/
  alarm(ALARM_DEFAULT);/*initial alarm*/
  if (sigemptyset(&criticalSectionSigMask_sigsetT) == -1){
    WriteError("Cannot prepare critical section\n");
    exit(1);
  }
  if (sigaddset(&criticalSectionSigMask_sigsetT, SIGALRM) == -1){
    WriteError("Cannot prepare critical section\n");
    exit(1);
  }
  while (numInterrogations_int < MAXINTERROGATIONS){

  	if (sigsetjmp(interrogation_sigjmpBuf, 1)){
	  /*alarm received, proceed*/
	  sendRequest_boolean = TRUE;
	  numInterrogations_int++;
	}
	else{
	  canjump_boolean = TRUE;
	  sendRequest_boolean = FALSE;
	}


	if(sendRequest_boolean|| firstRun_boolean){
		firstRun_boolean = FALSE;
	/*----Critical Section-------*/
		if (sigprocmask(SIG_BLOCK, &criticalSectionSigMask_sigsetT, &normalSigMask_sigsetT) == -1){
			WriteError("Cannot prepare critical section\n");
			exit(1);
		}

		if (forwardRequest() == -1){
		tearDownCommunication();
		WriteError("time client: Request could not be forwarded\n");
		exit(-1);
		}

		sprintf(outputMessage_sz, "Client %d: Request %d forwarded\n", client_pidT, numInterrogations_int);
		WriteError(outputMessage_sz);

		reply_sz = serverReply();
		if (reply_sz != NULL){
		  elapsedTimeInfo_sz = getTimeElapsed();
		  if (elapsedTimeInfo_sz != NULL){
		  	sprintf(outputMessage_sz, "client %d: time received: %s at %s\n", client_pidT, reply_sz, elapsedTimeInfo_sz);
		  	write(STDOUT_FILENO, outputMessage_sz, strlen(outputMessage_sz));
			free(elapsedTimeInfo_sz);
		  }
		  else WriteError("Cannot obtain client's accounting information\n");
		}
		else{
		tearDownCommunication();
		WriteError("time client: Reply could not be fetched\n");
		exit(-1);
		}
		alarm(ALARM_DEFAULT);/*reset alarm*/
		/*---- end of critical section---*/
		if (sigprocmask(SIG_UNBLOCK, &criticalSectionSigMask_sigsetT, &normalSigMask_sigsetT) == -1){
			WriteError("Cannot prepare critical section\n");
			exit(1);
		}
	}
	/*alarm(ALARM_DEFAULT);*/

	pause();


  }

  tearDownCommunication();
  return 0;
}


