#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include "timeserver_ipc.h"
#include "ipc_shared.h"
#include "../common/shared.h"

/*================== From man pages! Modify if not on linux*/
#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
       /* union semun is defined by including <sys/sem.h> */
#else
       /* according to X/OPEN we have to define it ourselves */
union semun {
   int val;                  /* value for SETVAL */
   struct semid_ds *buf;     /* buffer for IPC_STAT, IPC_SET */
   unsigned short *array;    /* array for GETALL, SETALL */
                                 /* Linux specific part: */
   struct seminfo *__buf;    /* buffer for IPC_INFO */
};
#endif


/*===================== IPC-Specific, private methods and data ===============*/

int semaphoreID = -1;
int sharedMemoryID = -1;
struct timeRequest * request_ptimeRequest = NULL;
struct timeRequest local_timeRequest;


/*Signal handler that terminates the process*/
static void SignalHandler(int sig){
#if defined(SIGNALCATCHING_TEST) || defined(IPC_EXCHANGE_TEST)
  printf("Handling signal and exiting\n");
#endif
  cleanupServerSetIPC();
  exit(0);
}

/*======================== Public methods ==========================*/
/*
function InitSignalHandler
task: Sets the signal handler to catch all killing signals
         The signal handler is meant as to ensure IPC stuctures are cleaned on termination
input: none
output: -1 on error, 0 otherwise
*/
int InitSignalHandler(){


	struct sigaction settings_sigaction;

	settings_sigaction.sa_flags = SA_NOCLDSTOP; /*Notify that we don't care for stopped child*/
	settings_sigaction.sa_handler = SignalHandler;

	/*Will ignore all signals when handler is on*/
	if (sigfillset(&settings_sigaction.sa_mask) == -1) return -1;

	return setSignalHandlerForTerminatingSignals(settings_sigaction);
}

/*
function initializedServerSetiPC
task: Initializes IPC structures
         The shared memory segment is mapped to SERVER_SHARED_MEMORY_KEY
	 The semaphore set is mapped to SERVER_SEMAPHORES_KEY
input; none
output: -1 on error, 0 otherwise
*/
int initializeServerSetIPC(){
  union semun initialize_semun;
  int i;

  /*Attach semaphore set*/
  semaphoreID = semget(SERVER_SEMAPHORES_KEY, /*key*/
                                         NUM_SERVER_SEMAPHORES, /*number of semaphores*/
					 IPC_CREATION_FLAG
                            );
  if (semaphoreID == -1) {
    WriteError("time server: Unable to get semaphore set\n");
    return -1; /*Error on semaphore creation*/
  }

  /*============ Initialize semaphores*/
  initialize_semun.val = 0;
  for(i = 0; i <NUM_SERVER_SEMAPHORES; i++){
    semctl(semaphoreID, i, SETVAL,initialize_semun);
  }


  /*Attach shared memory segment*/
  sharedMemoryID = shmget(SERVER_SHARED_MEMORY_KEY,
                                              REQUESTSIZE,
					      IPC_CREATION_FLAG
		);
  if (sharedMemoryID == -1){
    WriteError("time server: Unable to get shared memory segment\n");
    return -1;
  }

  request_ptimeRequest = shmat(sharedMemoryID,NULL, 0);
  if (request_ptimeRequest == (void *) -1){
    WriteError("time server: Unable to attach shared memory\n");
    return -1;
  }

  memset(request_ptimeRequest, 0,REQUESTSIZE);

  return 0;
}

/*
function cleanupServerSetIPC
task: Destroys IPC structures
input:none
output: none
*/
void cleanupServerSetIPC(){

  /*Clean semaphore*/
  if(semaphoreID != -1) {
    semctl(semaphoreID, 0, IPC_RMID, 0); /*Delete semaphore set*/
    semaphoreID = -1;
  }

  /*Clean shared memory segment*/
  if(sharedMemoryID != -1) {
   /*Detach memory segment*/
    shmdt(request_ptimeRequest);
    request_ptimeRequest  = NULL;

    /*remove shared memory segment from memory*/
    shmctl(sharedMemoryID, IPC_RMID, NULL);
    sharedMemoryID = -1;
  }

}
/*
function notifyServerReady
task: Sets semaphore SERVER_READY indicating server is ready
input: none
output: -1 on error, 0 otherwise
*/
int notifyServerReady(){
  struct sembuf init_sembuf;
  /*Signal server is ready*/
  init_sembuf.sem_num =SERVER_READY;
  init_sembuf.sem_op = +1;
  init_sembuf.sem_flg = 0;
  if (semop(semaphoreID, &init_sembuf, 1) == -1) return -1;

  return 0;
}

/*
function clientHasConnected
task: monitor semaphore SERVER_READY for client to obtain it
input: none
output: TRUE when client has connected
*/
boolean clientHasConnected(){
  int semaphoreValue_int;

  while(TRUE){
    semaphoreValue_int = semctl(semaphoreID, SERVER_READY,GETVAL);
    if(semaphoreValue_int > 0) sleepALittle(10);
    else {
      /*WriteError("Client notifies that request is in\n");*/
      break;
    }
  }

  return TRUE;
}
/*
function clientPutRequest
task: monitor semaphore REQUEST_FORWARDED_NOTIFICATION for some time to determine if client has put request in or not
input: none
output: returns TRUE when request is in, FALSE when request timed out
*/
boolean clientPutRequest(){
  struct sembuf lookup_sembuf;
/*
  const int MAXTIME = 10* 1000; / *10 seconds before timeout* /
  const int SLEEP = 10;/ *10 milliseconds* /
  const int NUMITERATIONS = MAXTIME/SLEEP;
  int i;

  lookup_sembuf.sem_num =REQUEST_FORWARDED_NOTIFICATION;
  lookup_sembuf.sem_op = -1;
  lookup_sembuf.sem_flg = IPC_NOWAIT;

  for (i = 0; i < NUMITERATIONS; i++){

	if (semop(semaphoreID, &lookup_sembuf, 1) == -1){
	  if (errno != EAGAIN) return FALSE;
	  sleepALittle(SLEEP);
	}
	else return TRUE;
  }

  return FALSE;
  */

  lookup_sembuf.sem_num =REQUEST_FORWARDED_NOTIFICATION;
  lookup_sembuf.sem_op = -1;
  lookup_sembuf.sem_flg = 0;

  if (semop(semaphoreID, &lookup_sembuf, 1) == -1) return FALSE;
  else return TRUE;

}
/*
function getRequest
task: obtain a request from the shared memory segment
input: none
return: return a pointer to a struct timeRequest holding the current request, NULL on error
*/
struct timeRequest * getRequest(){

  memcpy(&local_timeRequest,request_ptimeRequest, sizeof(local_timeRequest));

  /*return request_ptimeRequest;*/
  return &local_timeRequest;
}

/*====================== Test Drivers ==============================*/
#ifdef SIGNALCATCHING_TEST
#include <unistd.h>

int main(int argc, char** argv){
  InitSignalHandler();
  kill(getpid(),SIGINT);
  return 0;
}

#endif

#ifdef IPC_SET_TEST
#include <stdio.h>
#include <errno.h>
int main(int argc, char** argv){
  if (initializeServerSetIPC() == -1) return -1;
  system("ipcs");
  if (cleanupServerSetIPC() == -1){
    perror("Unable to clean up");
  }
  return 0;
}
#endif

#ifdef IPC_EXCHANGE_TEST
#include <errno.h>
#include <string.h>
#include "../common/serverlib.h"
int main(int argc, char** argv){
  struct sembuf test_sembuf;
  InitSignalHandler();
  initializeClock();

  WriteError(getCurrentTime());
  WriteError("\n");

  if (initializeServerSetIPC() == -1) return -1;

   /*try V()*/
  test_sembuf.sem_num =SERVER_READY;
  test_sembuf.sem_op = +1;
  test_sembuf.sem_flg = 0;

  if (semop(semaphoreID, &test_sembuf, 1) == -1){
    perror("Can't operate on semaphore");
    exit (-1);
  }

  /*try P()*/
  test_sembuf.sem_num =SERVER_READY;
  test_sembuf.sem_op = -1;
  test_sembuf.sem_flg = 0;

  if (semop(semaphoreID, &test_sembuf, 1) == -1){
    perror("Can't operate on semaphore");
    exit (-1);
  }

  if (clientPutRequest()){
    WriteError("Request received\n");
  }else WriteError("No request pending\n");

  cleanupServerSetIPC();

  WriteError(getCurrentTime());
  WriteError("\n");

  WriteError(getTimeElapsed());
  WriteError("\n");

 return 0;
}
#endif

