#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <setjmp.h>
int strcasecmp(const char *s1, const char *s2); /*From man pages, to avoid warning*/

#include "../common/shared.h"
#include "tcp_common.h"

int connectedSocket_int = -1; /*socket*/
volatile struct timeRequest our_timeRequest; /*The request to send periodiaclly*/
const int WAIT_TIME = 10; /*10s between server interrogations*/
char * PROMPT_SZ = "time> "; /*Prompt to display*/
int PROMPTSIZE_INT; /*size of the prompt*/
volatile boolean notTerminating_boolean = TRUE; /*if TRUE, we are in the disconnection protocol*/
volatile boolean canjump_boolean=FALSE; /*if TRUE, we can use siglongjmp, since sigsetjmp was set*/
sigjmp_buf disconnect_sigjmpBuf; /*used for the jumps*/
const int DISCONNECT_TIMEOUT = 30; /*timeout for the disconnection protocol*/
int logFileID_int= -1; /*FD of the log file*/

/*
function initializeLog
task: Open the file of the log
input: none
output: -1 on error, 0 otherwise
*/
int intializeLog(char * outputFileName_sz){
  mode_t log_modeT = 0600;
  logFileID_int = open(outputFileName_sz, O_WRONLY | O_CREAT |O_TRUNC | O_APPEND,log_modeT);
  if (logFileID_int <0) return -1;
  else return 0;
}

/*
function writeToLog
task: write a string to the log file
input: toLog_sz: zero-terminated string
output: -1 on error, 0 otherwise
*/
int writeToLog(char * toLog_sz){
  int logEntryLen_int = strlen(toLog_sz);
  int writtenLen_int;

  if (logFileID_int <0 || logEntryLen_int == 0) return -1;

  if ((writtenLen_int = write (logFileID_int, toLog_sz,logEntryLen_int)) != logEntryLen_int) return -1;
  return 0;
}

/*
function closeLog
task: close the log file
input: none
output: none
*/
void closeLog(){
  if (logFileID_int != -1) close (logFileID_int);
}

/*
function sendRequest
task: send a the current time request with data in network byte order
input: none
output: -1 on error, 0 otherwise
*/
int sendRequest(){
  struct timeRequest toSend_timeRequest;

  if(connectedSocket_int == -1) return -1;

  /*Move to network byte order*/
  toSend_timeRequest.reqType = htons(our_timeRequest.reqType);

  return send(connectedSocket_int,&toSend_timeRequest,sizeof(toSend_timeRequest), 0);

}

/*
function getReply
task: extract the reply from the socket in a non-blocking manner
input: reply_sz: zero-terminated string: string big enough to hold the reply (MAXREPLYSIZE)
output: -1: error, 0: no data, >0: data was read and is in reply_sz
*/
int getReply(char * reply_sz){
  struct pollfd socket_pollfd;

  if(connectedSocket_int == -1) return -1;

  socket_pollfd.fd =connectedSocket_int;
  socket_pollfd.events = POLLIN;

  if (poll(&socket_pollfd, 1,0) == -1) return -1;
  if (socket_pollfd.revents & POLLIN){ /*Data is in*/
    return recv(connectedSocket_int, reply_sz, MAXREPLYSIZE, 0);
  }

  return 0;
}

/*
function initConnection
task: initialize socket and connect to remote host
input: destination_sz: zero-terminated string: IP address of the destination host
output: -1 on error, 0 otherwise. See connect(2) for errno details
*/
int initConnection(const char * destination_sz){
  struct sockaddr_in connectAddress_sockaddrIn;

  connectedSocket_int = socket(AF_INET, SOCK_STREAM,0);
  if(connectedSocket_int == -1) return -1;

  if (setSocketParams(connectedSocket_int) == -1) return -1;

   /*Prepare bind-specific information From Donahoo/Calvert p.19*/
  memset(&connectAddress_sockaddrIn, 0, sizeof(struct sockaddr_in));
  connectAddress_sockaddrIn.sin_family = AF_INET;
  connectAddress_sockaddrIn.sin_addr.s_addr = inet_addr(destination_sz);
  connectAddress_sockaddrIn.sin_port = htons(SERVERPORT);

  return connect(connectedSocket_int,(struct sockaddr*)&connectAddress_sockaddrIn, sizeof(connectAddress_sockaddrIn));
}

/*
function setRequestType
task: verify new request type and notify user of new type, then set this information to the global request data (our_timeRequest)
input: reqInfo_int: int: type of request. Should be TIME, TIME_ELAPSED, DISCONNECT
output: -1 on invalid parameter, 0 otherwise
*/
int setRequestType(int reqInfo_int){

   switch (reqInfo_int){
   case TIME:
         WriteError("The server will now be queried for the time of the day\n");
	 writeToLog("\nThe server will now be queried for the time of the day\n");
   break;
   case TIME_ELAPSED:
     WriteError("The server will now be queried for the elapsed time\n");
     writeToLog("\nThe server will now be queried for the elapsed time\n");
   break;
   case DISCONNECT:
     WriteError("The server will be notified of disconnection\n");
     writeToLog("\nClient disconnecting\n");
   break;
   default:
     WriteError("Invalid request type\n");
     return -1;
   }

   our_timeRequest.reqType = reqInfo_int;
   return 0;
}

/*
function signalhandler
task: handle signals. SIGALRM will log data and send a request. Other signals will display an error to the user
input: sigType: signal type, as POSIX defines
output: none
*/
void signalHandler(int sigType){
  /*============================for alarm*/
  if(sigType == SIGALRM){
  /*Alarm set to send a request to the server*/
    if (notTerminating_boolean){
	if (sendRequest() != -1){
		switch (our_timeRequest.reqType){
		case TIME:
		writeToLog("\nQueried server for the time of the day\n");
		break;
		case TIME_ELAPSED:
		writeToLog("\nQueried server for the elapsed time\n");
		break;
		}
	} else writeToLog("\nCould not query server\n");
	alarm(WAIT_TIME);
    }
    else{ /*Alarm set for disconnection protocol error*/
      if (canjump_boolean){
        siglongjmp(disconnect_sigjmpBuf,1);
      }
    }
  }
  /*================= for other signals*/
  else{
    WriteError("Please use 'terminate' to quit the program\n");
  }
}

/*
function setSignalHandling
task: set alarm and terminating signals to be handled by function signalHandler
input: none
output: -1 on error, 0 otherwise
*/
int setSignalHandling(){
	struct sigaction settings_sigaction;

	settings_sigaction.sa_flags = SA_NOCLDSTOP; /*Notify that we don't care for stopped child*/
	settings_sigaction.sa_handler = signalHandler;
	/*Will ignore all signals when handler is on*/
	if (sigfillset(&settings_sigaction.sa_mask) == -1) return -1;

/*Set signal handling*/
	if(sigaction(SIGALRM,&settings_sigaction, NULL ) == -1) return -1;
	if (setSignalHandlerForTerminatingSignals(settings_sigaction) == -1) return -1;

	return 0;
}

/*
function printPrompt
task: display the prompt on the screen
input: none
output: see write(2)
*/
int printPrompt(){
  return write(STDOUT_FILENO,PROMPT_SZ,PROMPTSIZE_INT);
}


int main(int argc, char** argv){
  #define MAXINPUTLEN 16
  char replyBuf_sz[MAXREPLYSIZE];
  char toLog_sz[MAXREPLYSIZE+16];
  char inputCommand[MAXINPUTLEN];
  struct pollfd input_pollfd;
  int readData;
  char *newlineLoc_sz;

  /*============ argument validation*/
  if (argc != 3){
    WriteError("timeclient_tcp\nusage:\t timeclient_tcp <address of destination host> <name of log file>\n");
    exit(1);
  }

  /*=================== Intialize */
  if (setSignalHandling() == -1){
    WriteError("Cannot set signal handlers\n");
    exit(1);
  }

  if (intializeLog(argv[2]) == -1){
    WriteError("Cannot create specified log file\n");
    exit(1);
  }

  if (atexit(&closeLog) != 0){
    WriteError("Cannot set exit handler\n");
    exit(1);
  }
  if (initConnection(argv[1]) == -1){
    WriteError("Unable to connect to specified IP address\n");
    exit(1);
  }else WriteError("Client connected to server\n");


  setRequestType(TIME);
  /*WriteError("Request Type Set. The server will be queried regularly.\n");*/

  alarm(WAIT_TIME);

  memset(inputCommand, 0,MAXINPUTLEN);

  /*==========  Set prompt data */
  PROMPTSIZE_INT = strlen(PROMPT_SZ);

  /*Display prompt*/
  printPrompt();
  
  /*========================User input, reading from socket*/
  do{
    inputCommand[0] = '\0';

    /*=============== fetch from network*/
   readData = getReply(replyBuf_sz);
   if(readData < 0) {
     WriteError("Could not read from socket\n");
     exit(1);
   }
   else if (readData > 0){ /*received data, log it*/
     replyBuf_sz[readData] = '\0';
     sprintf(toLog_sz, "Reply received from server: %s\n", replyBuf_sz);
     writeToLog(toLog_sz);
   }
   else sleepALittle(20);

    /*=========read from user*/
    input_pollfd.fd =STDIN_FILENO;
    input_pollfd.events = POLLIN;
    if (poll(&input_pollfd, 1,0) == -1) return -1;

    if (input_pollfd.revents & POLLIN){ /*Data is in*/
      read(STDIN_FILENO, inputCommand, MAXINPUTLEN);
	/*Remove the ending '\n'*/
	if ((newlineLoc_sz = strchr(inputCommand,'\n')) != NULL){
	newlineLoc_sz[0] = '\0';
	}
	/*=============== examine user input*/
       if (strcasecmp(inputCommand, "TERMINATE") == 0){
         WriteError("Starting termination protocol.\n");
	 alarm(0);/*Cancelling alarms*/
         notTerminating_boolean = FALSE;
	 setRequestType(DISCONNECT);
       }
       else if (strcasecmp(inputCommand, "HELP") == 0){
         WriteError(
	   "Valid commands:\nhelp\t\t\tShows this list\nterminate\t\tDisconnect from server\ntime of day\t\tQueries the server for the time of the day\nelapsed time\t\tQueries the server for the time elapsed since it started running\n\n"
	 );
       }
       else if (strcasecmp(inputCommand, "TIME OF DAY") == 0){
         setRequestType(TIME);
       }
       else if (strcasecmp(inputCommand, "ELAPSED TIME") == 0){
         setRequestType(TIME_ELAPSED);
       }
       else {
         WriteError("Invalid input. Please type 'help' for valid commands\n");
       }

       /*==== re-display prompt*/
       if (notTerminating_boolean) printPrompt();
    }

  }while(notTerminating_boolean);

  /*============================ Client disconnecting =======================*/

  /*Set signal information*/
  if(sigsetjmp(disconnect_sigjmpBuf,1)){
    WriteError("\n\nError: server did not reply acknowledgment of disconnection\n\n");
    writeToLog("\n\nError: server did not reply acknowledgment of disconnection\n\n");
    exit(1);
  }

  alarm(DISCONNECT_TIMEOUT);
  /*Send disconnect request to server*/
  sendRequest();

  /*check if data is in. If so, say its ok and terminate*/
  while(TRUE){
    readData = getReply(replyBuf_sz);
    if(readData < 0) {
      WriteError("Could not read from socket\n");
      exit(1);
     }/*Received reply from server*/
     else if (readData > 0){
       alarm(0); /*Cancel alarm*/
       replyBuf_sz[readData] = '\0';
       sprintf(toLog_sz, "%s\n",replyBuf_sz);
       WriteError(toLog_sz);
       break;

     }
     else sleepALittle(20);
  }


  return 0;
}


